@@ -1,5 +1,5 @@
 /*
-* Copyright (c) 2013-2020, The PurpleI2P Project
+* Copyright (c) 2013-2021, The PurpleI2P Project
 *
 * This file is part of Purple i2pd project and licensed under BSD3
 *
@@ -574,9 +574,11 @@ namespace data
 						proxyReq.method = "CONNECT";
 						proxyReq.version = "HTTP/1.1";
 						proxyReq.uri = url.host + ":" + std::to_string(url.port);
+						auto auth = i2p::http::CreateBasicAuthorizationString (proxyUrl.user, proxyUrl.pass);
+						if (!auth.empty ())
+							proxyReq.AddHeader("Proxy-Authorization", auth);
 
 						boost::asio::streambuf writebuf, readbuf;
-
 						std::ostream out(&writebuf);
 						out << proxyReq.to_string();
 
@@ -676,8 +678,31 @@ namespace data
 			// direct connection
 			auto it = boost::asio::ip::tcp::resolver(service).resolve (
 				boost::asio::ip::tcp::resolver::query (url.host, std::to_string(url.port)), ecode);
-			if(!ecode)
-				s.lowest_layer().connect (*it, ecode);
+			if (!ecode)
+			{	
+				bool connected = false;
+				boost::asio::ip::tcp::resolver::iterator end;
+				while (it != end)
+				{	
+					boost::asio::ip::tcp::endpoint ep = *it;
+					if ((ep.address ().is_v4 () && i2p::context.SupportsV4 ()) ||
+					    (ep.address ().is_v6 () && i2p::context.SupportsV6 ()))
+					{	
+						s.lowest_layer().connect (ep, ecode);
+						if (!ecode)
+						{
+							connected = true;
+							break;
+						}	
+					}	
+					it++;
+				}
+				if (!connected)
+				{
+					LogPrint(eLogError, "Reseed: Failed to connect to ", url.host);
+					return "";
+				}	
+			}	
 		}
 		if (!ecode)
 		{
