@@ -28,7 +28,7 @@ namespace i2p
 
 	RouterContext::RouterContext ():
 		m_LastUpdateTime (0), m_AcceptsTunnels (true), m_IsFloodfill (false),
-		m_ShareRatio (100), m_Status (eRouterStatusUnknown),
+		m_ShareRatio (100), m_Status (eRouterStatusUnknown), m_StatusV6 (eRouterStatusUnknown),
 		m_Error (eRouterErrorNone), m_NetID (I2PD_NET_ID)
 	{
 	}
@@ -41,13 +41,15 @@ namespace i2p
 		if (!Load ())
 			CreateNewRouter ();
 		m_Decryptor = m_Keys.CreateDecryptor (nullptr);
+		m_TunnelDecryptor = m_Keys.CreateDecryptor (nullptr);
 		UpdateRouterInfo ();
 		if (IsECIES ())
-		{	
+		{
 			auto initState = new i2p::crypto::NoiseSymmetricState ();
 			i2p::crypto::InitNoiseNState (*initState, GetIdentity ()->GetEncryptionPublicKey ());
-			m_InitialNoiseState.reset (initState);	
-		}	
+			m_InitialNoiseState.reset (initState);
+			m_ECIESSession = std::make_shared<i2p::garlic::RouterIncomingRatchetSession>(*initState);
+		}
 	}
 
 	void RouterContext::CreateNewRouter ()
@@ -72,88 +74,100 @@ namespace i2p
 		bool ipv6;           i2p::config::GetOption("ipv6", ipv6);
 		bool ssu;            i2p::config::GetOption("ssu", ssu);
 		bool ntcp2;          i2p::config::GetOption("ntcp2.enabled", ntcp2);
-		bool ygg; 			 i2p::config::GetOption("meshnets.yggdrasil", ygg);
+		bool ygg;            i2p::config::GetOption("meshnets.yggdrasil", ygg);
 		bool nat;            i2p::config::GetOption("nat", nat);
-		std::string ifname;  i2p::config::GetOption("ifname", ifname);
-		std::string ifname4; i2p::config::GetOption("ifname4", ifname4);
-		std::string ifname6; i2p::config::GetOption("ifname6", ifname6);
-		uint8_t caps = 0;
+
+		if ((ntcp2 || ygg) && !m_NTCP2Keys)
+			NewNTCP2Keys ();
+		bool ntcp2Published = false;
+		if (ntcp2)
+		{
+			i2p::config::GetOption("ntcp2.published", ntcp2Published);
+			if (ntcp2Published)
+			{
+				std::string ntcp2proxy; i2p::config::GetOption("ntcp2.proxy", ntcp2proxy);
+				if (!ntcp2proxy.empty ()) ntcp2Published = false;
+			}
+		}
+		uint8_t caps = 0, addressCaps = 0;
 		if (ipv4)
 		{
 			std::string host = "127.0.0.1";
 			if (!i2p::config::IsDefault("host"))
 				i2p::config::GetOption("host", host);
-			else if (!nat && !ifname.empty())
-				/* bind to interface, we have no NAT so set external address too */
-				host = i2p::util::net::GetInterfaceAddress(ifname, false).to_string(); // v4
-
-			if(ifname4.size())
-				host = i2p::util::net::GetInterfaceAddress(ifname4, false).to_string();
+			else if (!nat)
+			{
+				// we have no NAT so set external address from local address
+				std::string address4; i2p::config::GetOption("address4", address4);
+				if (!address4.empty ()) host = address4;
+			}
 
+			if (ntcp2)
+			{
+				if (ntcp2Published)
+					routerInfo.AddNTCP2Address (m_NTCP2Keys->staticPublicKey, m_NTCP2Keys->iv, boost::asio::ip::address_v4::from_string (host), port);
+				else // add non-published NTCP2 address
+				{
+					addressCaps = i2p::data::RouterInfo::AddressCaps::eV4;
+					routerInfo.AddNTCP2Address (m_NTCP2Keys->staticPublicKey, m_NTCP2Keys->iv);
+				}
+			}
 			if (ssu)
-			{	
+			{
 				routerInfo.AddSSUAddress (host.c_str(), port, nullptr);
-				caps |= i2p::data::RouterInfo::eReachable | i2p::data::RouterInfo::eSSUTesting | i2p::data::RouterInfo::eSSUIntroducer; // R, BC
-			}	
+				caps |= i2p::data::RouterInfo::eReachable; // R
+			}
 		}
 		if (ipv6)
 		{
 			std::string host = "::1";
 			if (!i2p::config::IsDefault("host") && !ipv4) // override if v6 only
 				i2p::config::GetOption("host", host);
-			else if (!ifname.empty())
-				host = i2p::util::net::GetInterfaceAddress(ifname, true).to_string(); // v6
-
-			if(ifname6.size())
-				host = i2p::util::net::GetInterfaceAddress(ifname6, true).to_string();
-
-			if (ssu)
-			{	
-				routerInfo.AddSSUAddress (host.c_str(), port, nullptr);
-				caps |= i2p::data::RouterInfo::eReachable; // R
-			}	
-		}
-		
-		routerInfo.SetCaps (caps); // caps + L 
-		routerInfo.SetProperty ("netId", std::to_string (m_NetID));
-		routerInfo.SetProperty ("router.version", I2P_VERSION);
-		routerInfo.CreateBuffer (m_Keys);
-		m_RouterInfo.SetRouterIdentity (GetIdentity ());
-		m_RouterInfo.Update (routerInfo.GetBuffer (), routerInfo.GetBufferLen ());
-
-		if (ntcp2) // we don't store iv in the address if non published so we must update it from keys
-		{
-			if (!m_NTCP2Keys) NewNTCP2Keys ();
-			bool published; i2p::config::GetOption("ntcp2.published", published);
-			if (ipv4 || !published) UpdateNTCP2Address (true); // create not published NTCP2 address
-			if (published)
+			else
 			{
-				if (ipv4)
-					PublishNTCP2Address (port, true);
-				if (ipv6)
+				std::string address6; i2p::config::GetOption("address6", address6);
+				if (!address6.empty ()) host = address6;
+			}
+
+			if (ntcp2)
+			{
+				if (ntcp2Published)
 				{
-					// add NTCP2 ipv6 address
-					std::string host = "::1";
+					std::string ntcp2Host;
 					if (!i2p::config::IsDefault ("ntcp2.addressv6"))
-						i2p::config::GetOption ("ntcp2.addressv6", host);
-					m_RouterInfo.AddNTCP2Address (m_NTCP2Keys->staticPublicKey, m_NTCP2Keys->iv, boost::asio::ip::address_v6::from_string (host), port);
+						i2p::config::GetOption ("ntcp2.addressv6", ntcp2Host);
+					else
+						ntcp2Host = host;
+					routerInfo.AddNTCP2Address (m_NTCP2Keys->staticPublicKey, m_NTCP2Keys->iv, boost::asio::ip::address_v6::from_string (ntcp2Host), port);
+				}
+				else
+				{
+					if (!ipv4) // no other ntcp2 addresses yet
+						routerInfo.AddNTCP2Address (m_NTCP2Keys->staticPublicKey, m_NTCP2Keys->iv);
+					addressCaps |= i2p::data::RouterInfo::AddressCaps::eV6;
 				}
 			}
-			// enable added NTCP2 addresses
-			if (ipv4) m_RouterInfo.EnableV4 ();
-			if (ipv6) m_RouterInfo.EnableV6 ();
+			if (ssu)
+			{
+				routerInfo.AddSSUAddress (host.c_str(), port, nullptr);
+				caps |= i2p::data::RouterInfo::eReachable; // R
+			}
 		}
 		if (ygg)
 		{
 			auto yggaddr = i2p::util::net::GetYggdrasilAddress ();
 			if (!yggaddr.is_unspecified ())
-			{	
-				if (!m_NTCP2Keys) NewNTCP2Keys ();
-				m_RouterInfo.AddNTCP2Address (m_NTCP2Keys->staticPublicKey, m_NTCP2Keys->iv, yggaddr, port);
-				m_RouterInfo.EnableMesh ();
-				UpdateRouterInfo ();
-			}	
-		}	
+				routerInfo.AddNTCP2Address (m_NTCP2Keys->staticPublicKey, m_NTCP2Keys->iv, yggaddr, port);
+		}
+
+		if (addressCaps)
+			routerInfo.SetUnreachableAddressesTransportCaps (addressCaps);
+		routerInfo.SetCaps (caps); // caps + L
+		routerInfo.SetProperty ("netId", std::to_string (m_NetID));
+		routerInfo.SetProperty ("router.version", I2P_VERSION);
+		routerInfo.CreateBuffer (m_Keys);
+		m_RouterInfo.SetRouterIdentity (GetIdentity ());
+		m_RouterInfo.Update (routerInfo.GetBuffer (), routerInfo.GetBufferLen ());
 	}
 
 	void RouterContext::UpdateRouterInfo ()
@@ -185,10 +199,29 @@ namespace i2p
 			switch (m_Status)
 			{
 				case eRouterStatusOK:
-					SetReachable ();
+					SetReachable (true, false); // ipv4
 				break;
 				case eRouterStatusFirewalled:
-					SetUnreachable ();
+					SetUnreachable (true, false); // ipv4
+				break;
+				default:
+					;
+			}
+		}
+	}
+
+	void RouterContext::SetStatusV6 (RouterStatus status)
+	{
+		if (status != m_StatusV6)
+		{
+			m_StatusV6 = status;
+			switch (m_StatusV6)
+			{
+				case eRouterStatusOK:
+					SetReachable (false, true); // ipv6
+				break;
+				case eRouterStatusFirewalled:
+					SetUnreachable (false, true); // ipv6
 				break;
 				default:
 					;
@@ -211,25 +244,35 @@ namespace i2p
 			UpdateRouterInfo ();
 	}
 
-	void RouterContext::PublishNTCP2Address (int port, bool publish, bool v4only)
+	void RouterContext::PublishNTCP2Address (int port, bool publish, bool v4, bool v6, bool ygg)
 	{
 		if (!m_NTCP2Keys) return;
 		bool updated = false;
 		for (auto& address : m_RouterInfo.GetAddresses ())
 		{
-			if (address->IsNTCP2 () && (address->port != port || address->ntcp2->isPublished != publish) && (!v4only || address->host.is_v4 ()))
+			if (address->IsNTCP2 () && (address->port != port || address->published != publish))
 			{
-				if (!port && !address->port)
+				bool isAddr = v4 && address->IsV4 ();
+				if (!isAddr && (v6 || ygg))
 				{
-					// select random port only if address's port is not set
-					port = rand () % (30777 - 9111) + 9111; // I2P network ports range
-					if (port == 9150) port = 9151; // Tor browser
+					if (i2p::util::net::IsYggdrasilAddress (address->host))
+						isAddr = ygg;
+					else
+						isAddr = v6 && address->IsV6 ();
+				}
+				if (isAddr)
+				{
+					if (!port && !address->port)
+					{
+						// select random port only if address's port is not set
+						port = rand () % (30777 - 9111) + 9111; // I2P network ports range
+						if (port == 9150) port = 9151; // Tor browser
+					}
+					if (port) address->port = port;
+					address->published = publish;
+					address->ntcp2->iv = m_NTCP2Keys->iv;
+					updated = true;
 				}
-				if (port) address->port = port;
-				address->cost = publish ? 3 : 14;
-				address->ntcp2->isPublished = publish;
-				address->ntcp2->iv = m_NTCP2Keys->iv;
-				updated = true;
 			}
 		}
 		if (updated)
@@ -267,7 +310,7 @@ namespace i2p
 		bool updated = false;
 		for (auto& address : m_RouterInfo.GetAddresses ())
 		{
-			if (address->host != host && address->IsCompatible (host) && 
+			if (address->host != host && address->IsCompatible (host) &&
 			    !i2p::util::net::IsYggdrasilAddress (address->host))
 			{
 				address->host = host;
@@ -359,7 +402,7 @@ namespace i2p
 			case i2p::data::CAPS_FLAG_EXTRA_BANDWIDTH1 : limit = 2048; type = extra; break;
 			case i2p::data::CAPS_FLAG_EXTRA_BANDWIDTH2 : limit = 1000000; type = unlim; break; // 1Gbyte/s
 			default:
-				 limit =  48; type = low;
+				limit = 48; type = low;
 		}
 		/* update caps & flags in RI */
 		auto caps = m_RouterInfo.GetCaps ();
@@ -373,8 +416,8 @@ namespace i2p
 #if (__cplusplus >= 201703L) // C++ 17 or higher
 			[[fallthrough]];
 #endif
-			//  no break here, extra + high means 'X'
-			case high  : caps |= i2p::data::RouterInfo::eHighBandwidth;  break;
+			// no break here, extra + high means 'X'
+			case high : caps |= i2p::data::RouterInfo::eHighBandwidth; break;
 		}
 		m_RouterInfo.SetCaps (caps);
 		UpdateRouterInfo ();
@@ -390,6 +433,7 @@ namespace i2p
 		else if (limit >   48) { SetBandwidth('M'); }
 		else if (limit >   12) { SetBandwidth('L'); }
 		else                   { SetBandwidth('K'); }
+		m_BandwidthLimit = limit; // set precise limit
 	}
 
 	void RouterContext::SetShareRatio (int percents)
@@ -420,52 +464,60 @@ namespace i2p
 		}
 	}
 
-	void RouterContext::SetUnreachable ()
+	void RouterContext::SetUnreachable (bool v4, bool v6)
 	{
-		// set caps
-		uint8_t caps = m_RouterInfo.GetCaps ();
-		caps &= ~i2p::data::RouterInfo::eReachable;
-		caps |= i2p::data::RouterInfo::eUnreachable;
-		caps &= ~i2p::data::RouterInfo::eFloodfill;	// can't be floodfill
-		caps &= ~i2p::data::RouterInfo::eSSUIntroducer; // can't be introducer
-		m_RouterInfo.SetCaps (caps);
+		if (v4 || (v6 && !SupportsV4 ()))
+		{
+			// set caps
+			uint8_t caps = m_RouterInfo.GetCaps ();
+			caps &= ~i2p::data::RouterInfo::eReachable;
+			caps |= i2p::data::RouterInfo::eUnreachable;
+			caps &= ~i2p::data::RouterInfo::eFloodfill;	// can't be floodfill
+			m_RouterInfo.SetCaps (caps);
+		}
 		uint16_t port = 0;
 		// delete previous introducers
 		auto& addresses = m_RouterInfo.GetAddresses ();
 		for (auto& addr : addresses)
-			if (addr->ssu)
+			if (addr->ssu && ((v4 && addr->IsV4 ()) || (v6 && addr->IsV6 ())))
 			{
+				addr->published = false;
+				addr->caps &= ~i2p::data::RouterInfo::eSSUIntroducer; // can't be introducer
 				addr->ssu->introducers.clear ();
 				port = addr->port;
 			}
-		// remove NTCP2 v4 address
+		// unpiblish NTCP2 addreeses
 		bool ntcp2; i2p::config::GetOption("ntcp2.enabled", ntcp2);
 		if (ntcp2)
-			PublishNTCP2Address (port, false, true);
+			PublishNTCP2Address (port, false, v4, v6, false);
 		// update
 		UpdateRouterInfo ();
 	}
 
-	void RouterContext::SetReachable ()
+	void RouterContext::SetReachable (bool v4, bool v6)
 	{
-		// update caps
-		uint8_t caps = m_RouterInfo.GetCaps ();
-		caps &= ~i2p::data::RouterInfo::eUnreachable;
-		caps |= i2p::data::RouterInfo::eReachable;
-		caps |= i2p::data::RouterInfo::eSSUIntroducer;
-		if (m_IsFloodfill)
-			caps |= i2p::data::RouterInfo::eFloodfill;
-		m_RouterInfo.SetCaps (caps);
+		if (v4 || (v6 && !SupportsV4 ()))
+		{
+			// update caps
+			uint8_t caps = m_RouterInfo.GetCaps ();
+			caps &= ~i2p::data::RouterInfo::eUnreachable;
+			caps |= i2p::data::RouterInfo::eReachable;
+			if (m_IsFloodfill)
+				caps |= i2p::data::RouterInfo::eFloodfill;
+			m_RouterInfo.SetCaps (caps);
+		}
 		uint16_t port = 0;
 		// delete previous introducers
 		auto& addresses = m_RouterInfo.GetAddresses ();
 		for (auto& addr : addresses)
-			if (addr->ssu)
+			if (addr->ssu && ((v4 && addr->IsV4 ()) || (v6 && addr->IsV6 ())))
 			{
+				addr->published = true;
+				addr->caps |= i2p::data::RouterInfo::eSSUIntroducer;
 				addr->ssu->introducers.clear ();
 				port = addr->port;
 			}
-		// insert NTCP2 back
+		// publish NTCP2
 		bool ntcp2; i2p::config::GetOption("ntcp2.enabled", ntcp2);
 		if (ntcp2)
 		{
@@ -474,7 +526,7 @@ namespace i2p
 			{
 				uint16_t ntcp2Port; i2p::config::GetOption ("ntcp2.port", ntcp2Port);
 				if (!ntcp2Port) ntcp2Port = port;
-				PublishNTCP2Address (ntcp2Port, true, true);
+				PublishNTCP2Address (ntcp2Port, true, v4, v6, false);
 			}
 		}
 		// update
@@ -485,18 +537,17 @@ namespace i2p
 	{
 		if (supportsV6)
 		{
-			m_RouterInfo.EnableV6 ();
 			// insert v6 addresses if necessary
 			bool foundSSU = false, foundNTCP2 = false;
 			uint16_t port = 0;
 			auto& addresses = m_RouterInfo.GetAddresses ();
 			for (auto& addr: addresses)
 			{
-				if (addr->host.is_v6 ())
+				if (addr->IsV6 () && !i2p::util::net::IsYggdrasilAddress (addr->host))
 				{
 					if (addr->transportStyle == i2p::data::RouterInfo::eTransportSSU)
 						foundSSU = true;
-					else if (addr->IsPublishedNTCP2 ())
+					else if (addr->transportStyle == i2p::data::RouterInfo::eTransportNTCP)
 						foundNTCP2 = true;
 				}
 				port = addr->port;
@@ -529,6 +580,7 @@ namespace i2p
 					m_RouterInfo.AddNTCP2Address (m_NTCP2Keys->staticPublicKey, m_NTCP2Keys->iv, boost::asio::ip::address::from_string (ntcp2Host), ntcp2Port);
 				}
 			}
+			m_RouterInfo.EnableV6 ();
 		}
 		else
 			m_RouterInfo.DisableV6 ();
@@ -537,17 +589,63 @@ namespace i2p
 
 	void RouterContext::SetSupportsV4 (bool supportsV4)
 	{
+		// check if updates
+		if (supportsV4 && SupportsV4 ()) return;
+		if (!supportsV4 && !SupportsV4 ()) return;
+		// update
 		if (supportsV4)
+		{
+			bool foundSSU = false, foundNTCP2 = false;
+			std::string host = "127.0.0.1";
+			uint16_t port = 0;
+			auto& addresses = m_RouterInfo.GetAddresses ();
+			for (auto& addr: addresses)
+			{
+				if (addr->IsV4 ())
+				{
+					if (addr->transportStyle == i2p::data::RouterInfo::eTransportSSU)
+						foundSSU = true;
+					else if (addr->transportStyle == i2p::data::RouterInfo::eTransportNTCP)
+						foundNTCP2 = true;
+				}
+				if (addr->port) port = addr->port;
+			}
+			if (!port) i2p::config::GetOption("port", port);
+			// SSU
+			if (!foundSSU)
+			{
+				bool ssu; i2p::config::GetOption("ssu", ssu);
+				if (ssu)
+					m_RouterInfo.AddSSUAddress (host.c_str (), port, nullptr);
+			}
+			// NTCP2
+			if (!foundNTCP2)
+			{
+				bool ntcp2; i2p::config::GetOption("ntcp2.enabled", ntcp2);
+				if (ntcp2)
+				{
+					bool ntcp2Published; i2p::config::GetOption("ntcp2.published", ntcp2Published);
+					if (ntcp2Published)
+					{
+						uint16_t ntcp2Port; i2p::config::GetOption ("ntcp2.port", ntcp2Port);
+						if (!ntcp2Port) ntcp2Port = port;
+						m_RouterInfo.AddNTCP2Address (m_NTCP2Keys->staticPublicKey, m_NTCP2Keys->iv, boost::asio::ip::address::from_string (host), ntcp2Port);
+					}
+					else
+						m_RouterInfo.AddNTCP2Address (m_NTCP2Keys->staticPublicKey, m_NTCP2Keys->iv);
+				}
+			}
 			m_RouterInfo.EnableV4 ();
+		}
 		else
 			m_RouterInfo.DisableV4 ();
 		UpdateRouterInfo ();
 	}
 
 	void RouterContext::SetSupportsMesh (bool supportsmesh, const boost::asio::ip::address_v6& host)
-	{	
+	{
 		if (supportsmesh)
-		{	
+		{
 			m_RouterInfo.EnableMesh ();
 			uint16_t port = 0;
 			i2p::config::GetOption ("ntcp2.port", port);
@@ -561,16 +659,16 @@ namespace i2p
 				{
 					foundMesh = true;
 					break;
-				}	
+				}
 			}
 			if (!foundMesh)
 				m_RouterInfo.AddNTCP2Address (m_NTCP2Keys->staticPublicKey, m_NTCP2Keys->iv, host, port);
-		}	
+		}
 		else
 			m_RouterInfo.DisableMesh ();
 		UpdateRouterInfo ();
 	}
-		
+
 	void RouterContext::UpdateNTCP2V6Address (const boost::asio::ip::address& host)
 	{
 		bool isYgg = i2p::util::net::IsYggdrasilAddress (host);
@@ -638,7 +736,14 @@ namespace i2p
 			}
 		}
 		std::shared_ptr<const i2p::data::IdentityEx> oldIdentity;
-		if (m_Keys.GetPublic ()->GetSigningKeyType () == i2p::data::SIGNING_KEY_TYPE_DSA_SHA1)
+		bool rekey = m_Keys.GetPublic ()->GetSigningKeyType () == i2p::data::SIGNING_KEY_TYPE_DSA_SHA1;
+		if (!rekey && m_Keys.GetPublic ()->GetCryptoKeyType () == i2p::data::CRYPTO_KEY_TYPE_ELGAMAL)
+		{
+			// rekey routers with bandwidth = L (or default) this time
+			bool isFloodfill; i2p::config::GetOption("floodfill", isFloodfill);
+			if (!isFloodfill) rekey = true;
+		}
+		if (rekey)
 		{
 			// update keys
 			LogPrint (eLogInfo, "Router: router keys are obsolete. Creating new");
@@ -646,7 +751,7 @@ namespace i2p
 			m_Keys = i2p::data::PrivateKeys::CreateRandomKeys (i2p::data::SIGNING_KEY_TYPE_EDDSA_SHA512_ED25519,
 				i2p::data::CRYPTO_KEY_TYPE_ECIES_X25519_AEAD);
 			SaveKeys ();
-		}	
+		}
 		// read NTCP2 keys if available
 		std::ifstream n2k (i2p::fs::DataDirPath (NTCP2_KEYS), std::ifstream::in | std::ifstream::binary);
 		if (n2k)
@@ -669,8 +774,8 @@ namespace i2p
 			m_RouterInfo.Update (routerInfo.GetBuffer (), routerInfo.GetBufferLen ());
 			if (oldIdentity)
 				m_RouterInfo.SetRouterIdentity (GetIdentity ()); // from new keys
-			m_RouterInfo.SetProperty ("coreVersion", I2P_VERSION);
 			m_RouterInfo.SetProperty ("router.version", I2P_VERSION);
+			m_RouterInfo.DeleteProperty ("coreVersion"); // TODO: remove later
 		}
 		else
 		{
@@ -679,11 +784,11 @@ namespace i2p
 		}
 
 		if (IsUnreachable ())
-			SetReachable (); // we assume reachable until we discover firewall through peer tests
+			SetReachable (true, true); // we assume reachable until we discover firewall through peer tests
 
 		// read NTCP2
-		bool ntcp2;  i2p::config::GetOption("ntcp2.enabled", ntcp2);
-		bool ygg;  i2p::config::GetOption("meshnets.yggdrasil", ygg);
+		bool ntcp2; i2p::config::GetOption("ntcp2.enabled", ntcp2);
+		bool ygg; i2p::config::GetOption("meshnets.yggdrasil", ygg);
 		if (ntcp2 || ygg)
 		{
 			if (!m_NTCP2Keys) NewNTCP2Keys ();
@@ -716,15 +821,15 @@ namespace i2p
 		i2p::HandleI2NPMessage (CreateI2NPMessage (buf, GetI2NPMessageLength (buf, len)));
 	}
 
-	bool RouterContext::HandleCloveI2NPMessage (I2NPMessageType typeID, const uint8_t * payload, size_t len) 
-	{ 
+	bool RouterContext::HandleCloveI2NPMessage (I2NPMessageType typeID, const uint8_t * payload, size_t len)
+	{
 		auto msg = CreateI2NPMessage (typeID, payload, len);
 		if (!msg) return false;
 		i2p::HandleI2NPMessage (msg);
-		return true; 
-	} 
+		return true;
+	}
+
 
-		
 	void RouterContext::ProcessGarlicMessage (std::shared_ptr<I2NPMessage> msg)
 	{
 		std::unique_lock<std::mutex> l(m_GarlicMutex);
@@ -738,10 +843,12 @@ namespace i2p
 				return;
 			}
 			buf += 4;
-			auto session = std::make_shared<i2p::garlic::ECIESX25519AEADRatchetSession>(this, false);
-			session->HandleNextMessageForRouter (buf, len);
-		}	
-		else	
+			if (m_ECIESSession)
+				m_ECIESSession->HandleNextMessage (buf, len);
+			else
+				LogPrint (eLogError, "Router: Session is not set for ECIES router");
+		}
+		else
 			i2p::garlic::GarlicDestination::ProcessGarlicMessage (msg);
 	}
 
@@ -750,10 +857,10 @@ namespace i2p
 		if (i2p::data::netdb.GetPublishReplyToken () == bufbe32toh (msg->GetPayload () + DELIVERY_STATUS_MSGID_OFFSET))
 			i2p::data::netdb.PostI2NPMsg (msg);
 		else
-		{	
+		{
 			std::unique_lock<std::mutex> l(m_GarlicMutex);
 			i2p::garlic::GarlicDestination::ProcessDeliveryStatusMessage (msg);
-		}	
+		}
 	}
 
 	void RouterContext::CleanupDestination ()
@@ -772,36 +879,41 @@ namespace i2p
 		return m_Decryptor ? m_Decryptor->Decrypt (encrypted, data, ctx, true) : false;
 	}
 
-	bool RouterContext::DecryptTunnelBuildRecord (const uint8_t * encrypted, uint8_t * data, BN_CTX * ctx)
+	bool RouterContext::DecryptTunnelBuildRecord (const uint8_t * encrypted, uint8_t * data)
 	{
-		if (!m_Decryptor) return false;
+		if (!m_TunnelDecryptor) return false;
 		if (IsECIES ())
 		{
 			if (!m_InitialNoiseState) return false;
 			// m_InitialNoiseState is h = SHA256(h || hepk)
-			m_CurrentNoiseState.reset (new i2p::crypto::NoiseSymmetricState (*m_InitialNoiseState));		
+			m_CurrentNoiseState.reset (new i2p::crypto::NoiseSymmetricState (*m_InitialNoiseState));
 			m_CurrentNoiseState->MixHash (encrypted, 32); // h = SHA256(h || sepk)
 			uint8_t sharedSecret[32];
-			if (!m_Decryptor->Decrypt (encrypted, sharedSecret, ctx, false))
+			if (!m_TunnelDecryptor->Decrypt (encrypted, sharedSecret, nullptr, false))
 			{
 				LogPrint (eLogWarning, "Router: Incorrect ephemeral public key");
 				return false;
-			}	
-			m_CurrentNoiseState->MixKey (sharedSecret); 
+			}
+			m_CurrentNoiseState->MixKey (sharedSecret);
 			encrypted += 32;
 			uint8_t nonce[12];
 			memset (nonce, 0, 12);
-			if (!i2p::crypto::AEADChaCha20Poly1305 (encrypted, ECIES_BUILD_REQUEST_RECORD_CLEAR_TEXT_SIZE, 
+			if (!i2p::crypto::AEADChaCha20Poly1305 (encrypted, ECIES_BUILD_REQUEST_RECORD_CLEAR_TEXT_SIZE,
 				m_CurrentNoiseState->m_H, 32, m_CurrentNoiseState->m_CK + 32, nonce, data, ECIES_BUILD_REQUEST_RECORD_CLEAR_TEXT_SIZE, false)) // decrypt
 			{
 				LogPrint (eLogWarning, "Router: Tunnel record AEAD decryption failed");
 				return false;
-			}	
+			}
 			m_CurrentNoiseState->MixHash (encrypted, ECIES_BUILD_REQUEST_RECORD_CLEAR_TEXT_SIZE + 16); // h = SHA256(h || ciphertext)
 			return true;
-		}	
-		else	
-			return m_Decryptor->Decrypt (encrypted, data, ctx, false);
+		}
+		else
+		{
+			BN_CTX * ctx = BN_CTX_new ();
+			bool success = m_TunnelDecryptor->Decrypt (encrypted, data, ctx, false);
+			BN_CTX_free (ctx);
+			return success;
+		}
 	}
 
 	i2p::crypto::X25519Keys& RouterContext::GetStaticKeys ()
